//CCVisu is a tool for co-change visualization.
//This file is part of CCVisu. 

//Copyright (C) 2005  Dirk Beyer

//CCVisu is free software; you can redistribute it and/or
//modify it under the terms of the GNU Lesser General Public License
//as published by the Free Software Foundation; either
//version 2.1 of the License, or (at your option) any later version.

//CCVisu is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//Lesser General Public License for more details.

//You should have received a copy of the GNU Lesser General Public License
//along with CCVisu; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

//Please find the GNU Lesser General Public License in file
//../license_lgpl.txt or http://www.gnu.org/licenses/lgpl.txt

//Dirk Beyer    (firstname.lastname@epfl.ch)
//Ecole Polytechnique Fdrale de Lausanne (EPFL), Switzerland


package ccvisu;

import java.io.BufferedReader;
import java.util.Calendar;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.Vector;

/*****************************************************************
 * Reader for CVS log files. 
 * Extracts the co-change graph from the CVS log info.
 * @version  $Revision: 1.23 $; $Date: 2005/08/22 12:30:55 $
 * @author   Dirk Beyer
 *****************************************************************/
public class ReaderDataGraphCVS extends ReaderDataGraph {

  /** Time-window constant for transaction recovery, in milli-seconds.*/
  private int timeWindow;

  /**
   * Constructor.
   * @param in          Stream reader object.
   * @param timeWindow  Time window for transaction recovery, in milli-seconds
   *                    (default: 180'000).
   */
  public ReaderDataGraphCVS(BufferedReader in, int timeWindow) {
    super(in);
    this.timeWindow = timeWindow;
  }

  /** 
   * Represents a CVS revision entry (an abstraction of it).
   * @author Dirk Beyer
   */
  private class Revision {
    String filename;
    Long   time;
    String user;
    String logmsg;
    /** An artificially created number.*/
    int    transaction;
  };

  /*****************************************************************
   * Reads the edges of a graph in CVS log format
   * from stream reader <code>in</code>, 
   * and stores them in a list (of <code>GraphEdgeString</code> elements).
   * @return List of string edges.
   *****************************************************************/
  protected Vector readEdges() {
    Vector result = new Vector();

    Vector revisionList = readRevisionList();
    recoverTransactions(revisionList);
    
    for (int i = 0; i < revisionList.size(); ++i) {
      Revision revision = (Revision) revisionList.get(i);
      GraphEdgeString edge = new GraphEdgeString();
      // Source vertex.
      edge.x = Integer.toString(revision.transaction);
      // Target vertex.
      edge.y = revision.filename;
	  // Edge weight.
	  edge.w = "1.0";
      result.add(edge);
    }

    return result;
  }

  /*****************************************************************
   * Parses the date entry.
   * @param dateStr   The CVS date entry string.
   * @return  Long value of the date, or 
   *          <code>null</code> if <code>dateStr</code> 
   *          is not a valid date entry.
   *****************************************************************/
  private Long parseDate(String dateStr) {
    // Delimiter for year/month/day.
    char delim = '/';
    int posEnd = dateStr.indexOf(delim);
    if (posEnd < 0 || posEnd > 9) {
	  delim = '-';
	  posEnd = dateStr.indexOf(delim);
	  if (posEnd < 0 || posEnd > 9) {
	    return null;
	  }
    } 
    int posBegin = 0;
    int year = Integer.parseInt(dateStr.substring(posBegin, posEnd));
    posBegin = posEnd + 1;
    posEnd = dateStr.indexOf(delim, posBegin);
            
    int month = Integer.parseInt(dateStr.substring(posBegin, posEnd));
    posBegin = posEnd + 1;
    posEnd = dateStr.indexOf(' ', posBegin);
            
    int day = Integer.parseInt(dateStr.substring(posBegin, posEnd));
    posBegin = posEnd + 1;
    posEnd = dateStr.indexOf(':', posBegin);
            
    int hour = Integer.parseInt(dateStr.substring(posBegin, posEnd));
    posBegin = posEnd + 1;
    posEnd = dateStr.indexOf(':', posBegin);
            
    int min = Integer.parseInt(dateStr.substring(posBegin, posEnd));
    posBegin = posEnd + 1;
    posEnd = posBegin + 2;
    int sec =Integer.parseInt(dateStr.substring(posBegin, posEnd));
            
    Calendar cal = Calendar.getInstance();
    cal.clear();     // Erase the milli secs.
    cal.set(year, month-1, day, hour, min, sec);

    return new Long(cal.getTimeInMillis());
  }            

  /*****************************************************************
   * Parses the CVS log data and extracts revisions.
   * @return  List of revisions.
   *****************************************************************/
  private Vector readRevisionList() {
    Vector result = new Vector();

    String lLine = "";
    String filename = null;
    Long   time;
    String user;
    String logmsg;

    int lineno = 1;
    try {
      while ((lLine = in.readLine()) != null) {
		// New working file.
        if (lLine.startsWith("Working file: ")) {
          // Set name of the current working file, 
          //   for which we pasre the revisions.
          filename = lLine.substring(14);
        }
                  
		// New revision.
		if (lLine.startsWith("date: ")) {
		  // Set date, author, and logmsg of the current revision.

		  // Parse date. Start right after "date: ".
		  time = parseDate(lLine.substring(6, lLine.indexOf("author: ")));
		  if (time == null) {
			System.err.print("Error while reading the CVS date info for file: ");
			System.err.println(filename + ".");
		  }
            
		  // Parse author. Start right after "author: ".
		  int posBegin = lLine.indexOf("author: ") + 8;
		  int posEnd = lLine.indexOf(';', posBegin);
		  user = lLine.substring(posBegin, posEnd);
                  
		  // Parse logmsg. Start on next line the date/author line.
		  logmsg = "";
          ++lineno;
          while ( ((lLine = in.readLine()) != null)  &&
                   !lLine.startsWith("----")         &&
                   !lLine.startsWith("====")              ) {
            if (!lLine.startsWith("branches: ")) {
              logmsg += lLine + endl;
            }
            ++lineno;
          }

          // Create revision and add revision to resulting list.
          Revision revision = new Revision();
          revision.filename = filename;
          revision.time = time;
          revision.user = user;
          revision.logmsg = logmsg;
          result.add(revision);
          
          //System.out.print("File: " + filename + " Time: " + time.toString() +
          //                 " User: " + user + " LogMsg: " + logmsg);
          
        }
		++lineno;
      } // while
    } catch (Exception e) {
      System.err.println("Exception while reading the CVS log at line " 
                         + lineno + ":");
      System.err.println(e);
      System.err.print("Read line: ");
      System.err.println(lLine); 
    }
    return result;
  }

  /*****************************************************************
   * Recovers the change transactions for the co-change graph 
   *   from the revision information, i.e., it assignes
   *   the transaction ids for the revisions.
   * @param   revisionList is a list of revisions. IN/OUT parameter.
   *****************************************************************/
  private void recoverTransactions(Vector revisionList) {

    // Step 1: Transform the list of revisions to a sorted data structure.

    // A map user -> msg-entry.
    HashMap userMap = new HashMap();
    // A map logmsg -> time-entry.
    HashMap msgMap;
    // A map time -> file.
    TreeMap timeMap;
    // A set of files.
    TreeSet fileSet;
    
    for (int i = 0; i < revisionList.size(); ++i) {
      Revision revision = (Revision) revisionList.get(i);

      // A map logmsg -> time-entry.
      msgMap = (HashMap) userMap.get(revision.user);
      if (msgMap == null) {
        msgMap = new HashMap();
        userMap.put(revision.user, msgMap);
      }
      
        // A map time -> file.
        timeMap = (TreeMap) msgMap.get(revision.logmsg);
        if (timeMap == null) {
          timeMap = new TreeMap();
          msgMap.put(revision.logmsg, timeMap);
        }
        
          // A set of files.
          fileSet = (TreeSet) timeMap.get(revision.time);
          if (fileSet == null) {
            fileSet = new TreeSet();
            timeMap.put(revision.time, fileSet);
          }

            // Add file to set.
            fileSet.add(revision.filename);
    }


    // Step 2: Dump the info back into the list,
    //   with the additional entry 'transaction id'.

    revisionList.clear();

    int transaction = 0;
    Set userSet = userMap.keySet();
    Iterator userIt = userSet.iterator();
    while( userIt.hasNext() ) {
      String user = (String) userIt.next();
      msgMap = (HashMap) userMap.get(user);

      Set msgSet = msgMap.keySet();
      Iterator msgIt = msgSet.iterator();
      while( msgIt.hasNext() ) {
        String logmsg = (String) msgIt.next();
        timeMap = (TreeMap) msgMap.get(logmsg);

        Set timeSet = timeMap.keySet();
        Iterator timeIt = timeSet.iterator();

        long firstTime = 0;
        Set tmpFilesSeen = new TreeSet();  // Detect a time window that is too long.
        while( timeIt.hasNext() ) {
          Long time = (Long) timeIt.next();
          if (time.longValue() - firstTime > timeWindow) {
            ++transaction;
            firstTime = time.longValue();
            tmpFilesSeen.clear();
          }
          fileSet = (TreeSet) timeMap.get(time);

          Iterator fileIt = fileSet.iterator();
          while( fileIt.hasNext() ) {
            String filename = (String) fileIt.next();
			// Detect a time window that is too long.
			if (Verbosity.level >= 1 && tmpFilesSeen.contains(filename)) {
              System.err.println(
                  "Transaction-recovery warning: Time window might be to wide " + endl
                + "(currently '" + timeWindow + "' milli-seconds). " + endl
                + "File '" + filename + "' already contained in current transaction."
              );
            }
            tmpFilesSeen.add(filename);

            // Create revision and add revision to resulting list.
            Revision revision = new Revision();
            revision.filename = filename;
            revision.time = time;
            revision.user = user;
            revision.logmsg = logmsg;
            revision.transaction = transaction;
            revisionList.add(revision);
          }
        }
      }
    }

  }

};

