//CCVisu is a tool for co-change visualization.
//This file is part of CCVisu. 

//Copyright (C) 2005  Dirk Beyer

//CCVisu is free software; you can redistribute it and/or
//modify it under the terms of the GNU Lesser General Public License
//as published by the Free Software Foundation; either
//version 2.1 of the License, or (at your option) any later version.

//CCVisu is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//Lesser General Public License for more details.

//You should have received a copy of the GNU Lesser General Public License
//along with CCVisu; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

//Please find the GNU Lesser General Public License in file
//../license_lgpl.txt or http://www.gnu.org/licenses/lgpl.txt

//Dirk Beyer    (firstname.lastname@epfl.ch)
//Ecole Polytechnique Fdrale de Lausanne (EPFL), Switzerland


package ccvisu;

import java.awt.Color;

/*****************************************************************
 * Writer for graphical output of layout data.
 * @version  $Revision: 1.12 $; $Date: 2005/08/22 12:42:53 $
 * @author   Dirk Beyer
 *****************************************************************/
public abstract class WriterDataGraphics extends WriterData {

  protected boolean hideSource;
  protected float minVert;
  protected int fontSize;
  protected Color backColor;
  protected boolean blackCircle;

  /**
   * Constructor.
   * @param graph       Graph representation, contains the positions of the vertices.
   * @param hideSource  If true, draw only vertices that are not source of an edge.
   *                    In co-change graphs, all change-transaction vertices 
   *                    are source vertices.
   * @param minVert     Diameter of the smallest vertex.
   * @param fontSize    Font size of vertex annotations.
   * @param backColor   Background color.
   * @param blackCircle If true, draw black circle around each vertex.
   */
  public WriterDataGraphics(GraphData graph, 
                            boolean hideSource, 
                            float minVert,
	                        int fontSize,
	                        Color backColor,
	                        boolean blackCircle) {
    super(graph);
    this.hideSource = hideSource;
    this.minVert = minVert;
    this.fontSize = fontSize;
    this.backColor = backColor;
    this.blackCircle = blackCircle;
  }

  /*****************************************************************
   * Writes the layout data in a graphics format.
   *****************************************************************/
  abstract public void write();

  /*****************************************************************
   * Write graphics layout.
   * @param size  Size of output area (e.g., number of pixel).
   *****************************************************************/
  public void writeGraphicsLayout(int size) {
	float xPosMin = 0;
	float xPosMax = 0;
	float yPosMin = 0;
	float yPosMax = 0;
	float zPosMin = 0;
	float zPosMax = 0;
	for (int i = 0; i < graph.vertices.size(); ++i) {
	  if ( !(hideSource && ((GraphVertex) graph.vertices.get(i)).isSource) ) {
		xPosMin = Math.min(xPosMin, graph.pos[i][0]);
		xPosMax = Math.max(xPosMax, graph.pos[i][0]);
		yPosMin = Math.min(yPosMin, graph.pos[i][1]);
		yPosMax = Math.max(yPosMax, graph.pos[i][1]);
		zPosMin = Math.min(zPosMin, graph.pos[i][2]);
		zPosMax = Math.max(zPosMax, graph.pos[i][2]);
	  } 
	}

	float layoutDist;
	layoutDist = Math.max(xPosMax - xPosMin,  yPosMax - yPosMin);
	layoutDist = Math.max(layoutDist,         zPosMax - zPosMin);
	float xOffset = - xPosMin + 0.05f * layoutDist;
	float yOffset = - yPosMin + 0.05f * layoutDist;
	float zOffset = - zPosMin + 0.05f * layoutDist;
	float scale = 0.9f * size / layoutDist;

	// Draw the vertices.
	for (int i = 0; i < graph.vertices.size(); ++i) {
	  GraphVertex curVertex = (GraphVertex) graph.vertices.get(i);
	  if (!curVertex.showName) {
		if ( !(hideSource && curVertex.isSource) ) {
		  int dia  = (int) Math.max(Math.pow(curVertex.degree, 0.5) * minVert, 
									minVert);
		  int xPos = (int) ((graph.pos[i][0] + xOffset) *  scale);
		  int yPos = (int) ((graph.pos[i][1] + yOffset) * -scale + size);
		  int zPos = (int) ((graph.pos[i][2] + zOffset) *  scale);

		  writeVertex(curVertex, xPos, yPos, zPos, dia);
		}
	  }
	}

	// Draw the annotated vertices.
	for (int i = 0; i < graph.vertices.size(); ++i) {
	  GraphVertex curVertex = (GraphVertex) graph.vertices.get(i);
	  if (curVertex.showName) {
		if ( !(hideSource && curVertex.isSource) ) {
		  int dia  = (int) Math.max(Math.pow(curVertex.degree, 0.5) * minVert, 
									minVert);
		  int xPos = (int) ((graph.pos[i][0] + xOffset) *  scale);
		  int yPos = (int) ((graph.pos[i][1] + yOffset) * -scale + size);
		  int zPos = (int) ((graph.pos[i][2] + zOffset) *  scale);

		  writeVertex(curVertex, xPos, yPos, zPos, dia);
		}
	  }
	}

  }

  /**
   * Writes a vertex.
   * @param curVertex  The vertex object, to access vertex attributes.
   * @param xPos       x coordinate of the vertex.
   * @param yPos       y coordinate of the vertex.
   * @param zPos       z coordinate of the vertex.
   * @param dia        Diameter of the vertex.
   */
  abstract public void writeVertex(GraphVertex curVertex, 
                                   int xPos, int yPos, int zPos, int dia);

};
